import unittest
import recurly
import socket
from recurly import Resource, USER_AGENT
from recurly.resource import Empty
from .mock_resources import MyResource, MySubResource
from .mock_client import MockClient
import unittest.mock as mock
from unittest.mock import Mock, MagicMock
from datetime import datetime
from collections import OrderedDict
import sys


def delete_resource_client(request):
    conn = MagicMock()
    conn.request = request
    response = MagicMock()
    # empty response
    response.status = 204
    response.read.return_value = bytes("", "UTF-8")
    conn.getresponse = MagicMock(return_value=response)
    return mock.patch("http.client.HTTPSConnection", return_value=conn)


def update_resource_client(success, request):
    conn = MagicMock()
    conn.request = request
    response = MagicMock()
    if success:
        response.status = 201
        response.headers = {
            "Content-Type": "application/json",
            "X-Request-Id": "request-id",
        }
        response.read.return_value = bytes(
            """
            {
                "object": "my_resource", "my_int": 123
            }
            """,
            "UTF-8",
        )
    else:
        response.status = 422
        response.headers = {
            "Content-Type": "application/json",
            "X-Request-Id": "request-id",
        }
        response.read.return_value = bytes(
            """
            {
                "error": {
                    "type": "validation",
                    "message": "Year must be greater than 2000, Number is required",
                    "params":[{"param":"year","message":"must be greater than 2000"},{"param":"number","message":"is required"}]
                }
            }
            """,
            "UTF-8",
        )

    conn.getresponse = MagicMock(return_value=response)

    return mock.patch("http.client.HTTPSConnection", return_value=conn)


def get_socket_error_client():
    conn = MagicMock()
    conn.request = MagicMock(return_value=None)
    conn.getresponse = MagicMock(side_effect=socket.error("socket failure"))
    return mock.patch("http.client.HTTPSConnection", return_value=conn)


def get_resource_client(success, request):
    conn = MagicMock()
    conn.request = request
    response = MagicMock()
    if success:
        response.status = 200
        response.headers = {
            "Content-Type": "application/json",
            "X-Request-Id": "request-id",
        }
        response.read.return_value = bytes(
            """
            {
                "object": "my_resource", "my_int": 123
            }
            """,
            "UTF-8",
        )
    else:
        response.status = 404
        response.headers = {
            "Content-Type": "application/json",
            "X-Request-Id": "request-id",
        }
        response.read.return_value = bytes(
            """
            {
                "error": {
                   "type": "not_found",
                   "message": "Couldn't find Resource with id = 123",
                   "params": [{"param": "resource_id"}]
               }
            }
            """,
            "UTF-8",
        )

    conn.getresponse = MagicMock(return_value=response)

    return mock.patch("http.client.HTTPSConnection", return_value=conn)


expected_headers = {
    "Authorization": "Basic YXBpa2V5Og==",
    "Accept": "application/vnd.recurly.v2018-08-09",
    "Content-Type": "application/json",
    "User-Agent": USER_AGENT,
}


class TestBaseClient(unittest.TestCase):
    def test_api_version(self):
        client = MockClient("apikey")
        self.assertEqual(client.api_version(), "v2018-08-09")

    def test_pathParameterValidationEmpty(self):
        request = MagicMock(return_value=None)
        with get_resource_client(False, request) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.ApiError) as e:
                resource = client.get_resource("")

            self.assertEqual("Parameters cannot be empty strings", str(e.exception))

    def test_pathParameterValidationNone(self):
        request = MagicMock(return_value=None)
        with get_resource_client(False, request) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.ApiError) as e:
                resource = client.get_resource(None)

            self.assertEqual("Invalid parameter type", str(e.exception))

    def test_optionsValidationSuccess(self):
        request = MagicMock(return_value=None)
        with get_resource_client(True, request) as conn:
            client = MockClient("apikey")
            resource = client.get_resource(
                "123", params={"q": 123}, headers={"AcceptLanguage": "en"}
            )
            self.assertEqual(type(resource), MyResource)

    def test_optionsValidationFailure(self):
        request = MagicMock(return_value=None)
        with get_resource_client(True, request) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.ApiError) as e:
                resource = client.get_resource(
                    "123", params={"q": 123}, invalid="invalid"
                )

            self.assertEqual(
                "Invalid options: invalid. Allowed options: body, params, headers",
                str(e.exception),
            )

    def test_successful_GET_200(self):
        request = MagicMock(return_value=None)
        with get_resource_client(True, request) as conn:
            client = MockClient("apikey")
            resource = client.get_resource(
                "123", params={"q": 123}, headers={"Accept": "cannot override"}
            )
            request.assert_called_with(
                "GET", "/resources/123?q=123", None, headers=expected_headers
            )
            self.assertEqual(type(resource), MyResource)

    def test_failure_GET_404(self):
        request = MagicMock(return_value=None)
        with get_resource_client(False, request) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.errors.NotFoundError) as e:
                resource = client.get_resource("123")

            request.assert_called_with(
                "GET", "/resources/123", None, headers=expected_headers
            )
            err = e.exception.error
            self.assertEqual(err.type, "not_found")

    def test_successful_PUT_201(self):
        request = MagicMock(return_value=None)
        with update_resource_client(True, request) as conn:
            client = MockClient("apikey")
            resource = client.update_resource("123", {"my_int": 123})
            request.assert_called_with(
                "PUT", "/resources/123", """{"my_int": 123}""", headers=expected_headers
            )
            self.assertEqual(type(resource), MyResource)
            self.assertEqual(resource.my_int, 123)

    def test_failure_PUT_422(self):
        request = MagicMock(return_value=None)
        with update_resource_client(False, request) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.errors.ValidationError) as e:
                resource = client.update_resource("123", {"my_int": 123})

            request.assert_called_with(
                "PUT", "/resources/123", """{"my_int": 123}""", headers=expected_headers
            )
            err = e.exception.error
            self.assertEqual(err.type, "validation")

    def test_DELETE_204(self):
        request = MagicMock(return_value=None)
        with delete_resource_client(request) as conn:
            client = MockClient("apikey")
            resource = client.delete_resource("123")
            request.assert_called_with(
                "DELETE", "/resources/123", None, headers=expected_headers
            )
            self.assertIsInstance(resource, Empty)

    def test_datetimes_in_bodies_are_converted(self):
        request = MagicMock(return_value=None)
        with update_resource_client(True, request) as conn:
            client = MockClient("apikey")
            resource = client.update_resource("123", {"my_dt": datetime(2020, 1, 1)})
            request.assert_called_with(
                "PUT",
                "/resources/123",
                """{"my_dt": "2020-01-01T00:00:00"}""",
                headers=expected_headers,
            )

    def test_failure_socket_error(self):
        with get_socket_error_client() as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.NetworkError) as e:
                resource = client.update_resource("123", {"my_int": 123})

    def test_pathParameterEncoding(self):
        client = MockClient("apikey")
        path = client._interpolate_path("/resource/%s", "asdf/,123")
        self.assertEqual(path, "/resource/asdf%2F%2C123")

    def test_query_param_encoding(self):
        # This test does not work in python 3.5
        # TODO remove the if statement when we drop 3.5 support
        if sys.version_info[1] > 5:
            request = MagicMock(return_value=None)
            with get_resource_client(True, request) as conn:
                client = MockClient("apikey")
                d = datetime(2020, 10, 5)
                params = {"q": 123, "d": d, "c": False, "b": True}
                resource = client.get_resource("123", params=params)
                url = "/resources/123?q=123&d=2020-10-05T00%3A00%3A00&c=false&b=true"

                request.assert_called_with("GET", url, None, headers=expected_headers)
                self.assertEqual(type(resource), MyResource)

    def test_client_can_set_timeout(self):
        timeout = 3
        client = MockClient("apikey", timeout=timeout)
        self.assertEqual(client.__dict__["_BaseClient__conn"].timeout, timeout)



import recurly


class RecurlyError(Exception):
    @classmethod
    def error_from_status(cls, status):
        return recurly.errors.ERROR_MAP.get(status, "")


class ApiError(RecurlyError):
    def __init__(self, message, error):
        super(ApiError, self).__init__(message)
        self.error = error


class NetworkError(RecurlyError):
    pass



import unittest
import recurly
import platform
import json
from datetime import datetime
from datetime import timezone
from recurly import Resource, Response, Request
from pydoc import locate
from .mock_resources import MyResource, MySubResource
from unittest.mock import Mock, MagicMock

major, minor, patch = platform.python_version_tuple()


def cast(obj, class_name=None, resp=None):
    return Resource.cast_json(obj, class_name, resp)


def cast_error(obj, resp=None):
    return Resource.cast_error(obj)


class TestResource(unittest.TestCase):
    def test_cast_object_unknown_class(self):
        # should return the original dict
        obj = {"object": "unknown_class", "prop1": 1}
        # TODO test non-strict-mode behavior
        with self.assertRaises(ValueError):
            self.assertEqual(cast(obj), obj)

    def test_cast_page(self):
        # should return a page of cast data
        page = cast(
            {
                "object": "list",
                "has_more": True,
                "next": "/resources?cursor=123",
                "data": [
                    {"object": "my_resource", "my_string": "kmxu3f3qof17"},
                    {"object": "my_resource", "my_string": "kmxu3f3qof18"},
                ],
            }
        )

        self.assertEqual(type(page), recurly.resource.Page)
        self.assertEqual(page.has_more, True)
        self.assertEqual(page.next, "/resources?cursor=123")
        self.assertEqual(type(page.data[0]), MyResource)
        self.assertEqual(page.data[0].my_string, "kmxu3f3qof17")

    def test_cast_non_json_error_0(self):
        resp = MagicMock()
        resp.request_id = "1234"
        resp.status = 0
        err = cast_error(resp)

        # When the error class is unknown, it should fallback to ApiError
        self.assertEqual(type(err), recurly.ApiError)
        self.assertEqual(str(err), "Unexpected 0 Error. Recurly Request Id: 1234")

    def test_cast_non_json_error_500(self):
        resp = MagicMock()
        resp.request_id = "1234"
        resp.status = 500
        err = cast_error(resp)

        # Error class should be determined from HTTP status code when not json
        self.assertEqual(type(err), recurly.errors.InternalServerError)
        self.assertEqual(str(err), "Unexpected 500 Error. Recurly Request Id: 1234")

    def test_cast_unknown_error(self):
        resp = MagicMock()
        resp.request_id = "1234"
        resp.content_type = "application/json"
        resp.body = json.dumps(
            {"error": {"type": "unknown", "message": "Error Message"}}
        ).encode("UTF-8")
        err = cast_error(resp)

        # When the error class is unknown, it should fallback to ApiError
        self.assertEqual(type(err), recurly.ApiError)
        self.assertEqual(str(err), "Error Message. Recurly Request Id: 1234")

    def test_cast_error(self):
        resp = MagicMock()
        resp.request_id = "1234"
        resp.content_type = "application/json"
        resp.body = json.dumps(
            {"error": {"type": "validation", "message": "Invalid"}}
        ).encode("UTF-8")
        err = cast_error(resp)

        self.assertEqual(type(err), recurly.errors.ValidationError)
        self.assertEqual(str(err), "Invalid. Recurly Request Id: 1234")

    def test_cast(self):
        obj = cast(
            {
                "object": "my_resource",
                "my_string": "string",
                "my_int": 123,
                "my_float": 1.123,
                "my_bool": False,
                "my_datetime": "2022-01-01T00:00:00Z",
                "my_sub_resource": {"object": "my_sub_resource", "my_string": "string"},
                "my_sub_resources": [
                    {"object": "my_sub_resource", "my_string": "string1"},
                    {"object": "my_sub_resource", "my_string": "string2"},
                ],
            }
        )

        # First check that all types and sub types were casts
        self.assertEqual(type(obj), MyResource)
        self.assertEqual(type(obj.my_sub_resource), MySubResource)
        self.assertEqual(type(obj.my_sub_resources[0]), MySubResource)
        self.assertEqual(type(obj.my_sub_resources[1]), MySubResource)

        # Check primitive types
        self.assertEqual(obj.my_string, "string")
        self.assertEqual(obj.my_int, 123)
        self.assertEqual(obj.my_float, 1.123)
        self.assertEqual(obj.my_bool, False)
        if major >= "3" and minor >= "7":
            self.assertEqual(
                obj.my_datetime, datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
            )
            self.assertEqual(obj.my_datetime.tzinfo, timezone.utc)
        else:
            self.assertEqual(obj.my_datetime, datetime(2022, 1, 1, 0, 0, 0))
        self.assertEqual(obj.my_sub_resource.my_string, "string")
        self.assertEqual(obj.my_sub_resources[0].my_string, "string1")
        self.assertEqual(obj.my_sub_resources[1].my_string, "string2")

    def test_repr_str(self):
        resource = cast({"object": "my_resource", "my_string": "kmxu3f3qof17"})
        # Should return the string version of vars(resource)
        self.assertIn("'object': 'my_resource'", repr(resource))
        self.assertIn("'my_string': 'kmxu3f3qof17'", repr(resource))
        # str() should default to __repr__() implementation
        self.assertIn("'object': 'my_resource'", str(resource))
        self.assertIn("'my_string': 'kmxu3f3qof17'", str(resource))
